@@ -8,39 +8,39 @@ like py.test or nose
 """
 
 __author__ = 'Steven LI'
+__version__ = '0.8.6'
 
 import logging
 import os, re, time
 from inspect import currentframe
 import operator
+from .ymal_testbed import init_testbed
 
 __all__ = ['test_logger', 'ok', 'fail', 'eq', 'ne', 'gt', 'lt', 'le', 'ge', 'match', 'unmatch',
+           'has', 'hasnt',
            'setlogger', 'addBiOperator', 'getOpWrapper', 'step', 'steps', 's', 'check', 'checks',
-           'addStepOption' ]
-
-__tracebackhide__ = True
-
+           'addStepOption', 'log_new_func', 'auto_func_detection',
+           'ReturnPassSet', 'init_testbed']
 
 def __init_logger__():
     global test_logger
 
-    file_name = time.strftime('/tmp/test'+"_%Y_%m_%d_%H_%M_%S.log")
-    fh = logging.FileHandler(file_name)
-    ch = logging.StreamHandler()
+    ##### Remove the following handler to provide more flexibility
+    # file_name = time.strftime('/tmp/test'+"_%Y%m%d_%H%M.log")
+    # fh = logging.FileHandler(file_name)
+    # formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
+    # fh.setFormatter(formatter)
+    # test_logger.addHandler(fh)
 
-    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
-    fh.setFormatter(formatter)
+    ch = logging.StreamHandler()
     formatter = logging.Formatter('%(levelname)s - %(message)s')
     ch.setFormatter(formatter)
-
-    test_logger.addHandler(fh)
     test_logger.addHandler(ch)
 
 
 test_logger = logging.getLogger("Test")
 test_logger.setLevel(logging.DEBUG)
 
-
 def setlogger(testlogger):
     global test_logger
     test_logger = testlogger
@@ -68,6 +68,16 @@ class __TestLog__(object):
         self.case_name = co.co_name
         self.file_name = co.co_filename
         self.case_no += 1
+        self.__log_case__()
+
+    def new_func(self, name, path):
+        self.step_no = 0
+        self.case_name = name
+        self.file_name = path
+        self.case_no += 1
+        self.__log_case__()
+
+    def __log_case__(self):
         self.step_logger.info("------------------------------------------------------")
         self.step_logger.info("Func %s in file: %s" % (self.case_name, self.file_name))
 
@@ -82,6 +92,20 @@ class __TestLog__(object):
 _this_file = os.path.normcase(setlogger.__code__.co_filename)
 __step_info__ = __TestLog__()
 __init_logger__()
+___auto_func_detection___ = True
+
+
+def log_new_func(name=None, path=None):
+    if name:
+        __step_info__.new_func(name, path)
+    else:
+        co = _invoker().f_code
+        __step_info__.new_case(co)
+
+
+def auto_func_detection(auto=True):
+    global ___auto_func_detection___
+    ___auto_func_detection___ = auto
 
 
 def _invoker():
@@ -97,28 +121,22 @@ def _invoker():
 
 
 def _step_closure(func):
-    """
-    decorator of closure: execution and logging for step functions
+    """decorator of closure: execution and logging for step functions
+
     :param func: the step function
     :return: True, or raise exception if error
     """
-    __tracebackhide__ = True
 
     def __step__(*args, **kwargs):
         __tracebackhide__ = True
-        # Get current the caller of this function outside this file
-        # f = currentframe().f_back
-        # while hasattr(f, "f_code"):
-        #     co = f.f_code
-        #     filename = os.path.normcase(co.co_filename)
-        #     if filename == _this_file:
-        #         f = f.f_back
-        #         continue
 
-        co = _invoker().f_code
-        if co != __step_info__.case_obj:
-            __step_info__.new_case(co)
-        #break
+        if ___auto_func_detection___:
+            f = _invoker()
+            while '__auto_func_detection__' in f.f_locals and not f.f_locals['__auto_func_detection__']:
+                f = f.f_back
+            co = f.f_code
+            if co != __step_info__.case_obj:
+                __step_info__.new_case(co)
 
         # Get the step message
         (pf, step_info, err_msg) = func(*args, **kwargs)
@@ -130,6 +148,13 @@ def _step_closure(func):
     return (__step__)
 
 
+##
+ReturnPassSet = {0, None}
+def PassedOrNot(cond):
+    for v in ReturnPassSet:
+        if cond is v: return True
+    return bool(cond)
+
 @_step_closure
 def __ok__(cond, desc, errmsg):
     __tracebackhide__ = True
@@ -138,11 +163,12 @@ def __ok__(cond, desc, errmsg):
 
 @_step_closure
 def ok(cond, passdesc=None, faildesc=None):
-    '''
+    """ One test step to be logged, pass if cond
     :param cond: could be a string, when there is no desc parameter and just pass the step
-    :param desc: description of this step
+    :param passdesc: description of this step to be logged if passed
+    :param faildesc: description of this step to be logged if failed
     :return: True when it passed
-    '''
+    """
     if not passdesc:
         if isinstance(cond, bool):
             passdesc = "ok(%r)" % cond
@@ -151,7 +177,7 @@ def ok(cond, passdesc=None, faildesc=None):
             cond = True
     if not faildesc: faildesc = passdesc
 
-    if cond:
+    if PassedOrNot(cond):
         return (True, passdesc, '')
     else:
         return (False, faildesc, '')
@@ -188,7 +214,7 @@ class TestStepFail(Exception):
     """ custom exception for error reporting. """
 
 
-class TestRunTimeError(Exception):
+class TestRunTimeError(RuntimeError):
     """ custom exception for error reporting. """
 
 
@@ -220,13 +246,17 @@ getOpWrapper = lambda op_string: _bi_comp_closure(op_string)
 
 
 ## To add the =~ and !~ operator
-# match = BiOpRegister('=~', lambda o1, o2: re.compile(o2).match(o1))
-# unmatch = BiOpRegister( '!~', lambda o1, o2: not re.compile(o2).match(o1) )
 addBiOperator('=~', lambda o1, o2: re.compile(o2).search(o1))
 match = getOpWrapper('=~')
 addBiOperator('!~', lambda o1, o2: not re.compile(o2).search(o1))
 unmatch = getOpWrapper('!~')
 
+## To add =<(has) and !< (hasnt) operator
+addBiOperator('=>', lambda o1, o2: o2 in o1)
+has = getOpWrapper('=>')
+addBiOperator('!>', lambda o1, o2: not o2 in o1)
+hasnt = getOpWrapper('!>')
+
 
 ####################################################################################
 ## The section is for step functions implementation
@@ -256,10 +286,7 @@ def step(code_string, globals=None, locals=None, **kwargs):
 def steps(code_lines, globals=None, locals=None, batch=False):
     __tracebackhide__ = True
     step_list = code_lines.split('\n')
-    # if not globals:
-    #     invoker = currentframe().f_back
-    #     globals = invoker.f_globals
-    #     locals = invoker.f_locals
+
     together = False
     half = ''
     line_no = 0
@@ -286,37 +313,31 @@ def steps(code_lines, globals=None, locals=None, batch=False):
                 step_res = step(code_string, globals, locals, **options)
             except Exception as e:
                 if not batch: raise e
-                test_logger.debug("%d: %s - FAIL - %r" % (line_no, code_string, e))
+                test_logger.debug("    ^^^ %d: %s - FAIL - %r   ^^^" % (line_no, code_string, e))
                 failed_steps.append({'line': line_no,
                                      'step': step_no,
                                      'code': code_string,
                                      'exception': e})
+                step_res = {'exception': e,
+                            'code': code_string,
+                            'info': e.args,
+                            'result': False}
         step_results.append(step_res)
 
     if len(failed_steps):
-        __ok__(False, '%d checks failed' % len(failed_steps),
-               '1st failed step: %d %s' % (line_no, failed_steps[0]['code']))
+        __ok__(False, 'Overall Batch Result: %d checks failed' % len(failed_steps),
+               '1st failed step: (line_%d) %s\nerr_msg: %s' % (line_no, failed_steps[0]['code']) )
     return {'result': True,
             'step_results': step_results}
 
 
-s = steps
-
 
 class TestStep:
     def __init__(self, code_string, globals, locals, **kwargs):
         self.code_string = code_string
         (self.globals, self.locals) = (globals, locals)
         self.kwargs = kwargs
-        # self.options = {
-        #     # format: optionString: [HasThisOption, parameter, func]
-        #     'xfail': [False, False],
-        #     'repeat': [False,  0],
-        #     'timeout': [False, 30],
-        #     'duration': [False, 30],
-        #     'warning': [False, False],
-        #     'skip': [False, False]
-        # }
+
         self.options = dict((k, [False, None]) for k in TestStepOptions.keys())
 
         self.op_string = None
@@ -327,10 +348,12 @@ class TestStep:
         self.pass_str = self.fail_str = code_string
         self.result = False
         self.exception = None
+        self.warn_msg = None
+        self.debug_msg = None
 
     @classmethod
     def parse_steps(cls, step_string):
-        pattern = re.compile(r'(.*)\s+(?=(?:-\w(?:\s|$)|--\w{2,}(?:\s|$)))(.*)')
+        pattern = re.compile(r'(.*?)\s+(?=(?:-\w(?:\s|$)|--\w{2,}(?:\s|$)))(.*)')
         m = pattern.match(step_string)
         code_string = step_string
         option_d = {}
@@ -374,11 +397,6 @@ class TestStep:
         for (k, v) in self.kwargs.items():
             if k in self.options.keys():
                 self.options[k][0] = True
-                # paramType = TestStep.options[k][1]
-                # if paramType == 'int':
-                #     self.options[k][2] = int(v)
-                # elif paramType == 'bool':
-                #     self.options[k][2] = bool(re.compile('T|t|Y|y').match(v))
                 self.options[k][1] = v
             else:  # should not happen
                 raise RuntimeError('ParameterType Error for option: %s' % k)
@@ -394,8 +412,10 @@ class TestStep:
         else:
             self.err_msg = "%r" % (self.expr1_val)
 
-        self.result = bool(ret)
-        return ret
+        #self.result = bool(ret)
+        #return ret
+        self.result = PassedOrNot(ret)
+        return self.result
 
     def execute(self):
         __tracebackhide__ = True
@@ -407,6 +427,10 @@ class TestStep:
         self.result = self.func()
         step_description = self.pass_str if self.result else self.fail_str
         __ok__(self.result, step_description, self.err_msg)
+        if self.warn_msg:
+            test_logger.warning(self.warn_msg)
+        if self.debug_msg:
+            test_logger.debug(self.debug_msg)
 
     @classmethod
     def _repeat(cls, obj, seconds):
@@ -424,7 +448,8 @@ class TestStep:
                     time.sleep(1)
                 obj.err_msg += ' - tried %d times in %d seconds' % (loop, seconds)
                 obj.result = bool(p_f)
-                test_logger.debug("Results(-r %d set) { %s}" % (seconds, debug_info))
+                #obj.debug_msg = "Results(-r %d set) { %s}" % (seconds, debug_info)
+                test_logger.debug("   vvv  Results(-r %d set) { %s}  vvv" % (seconds, debug_info))
                 return p_f
 
             return do_it
@@ -443,7 +468,7 @@ class TestStep:
                 t.start()
                 t.join(seconds)
                 if t.is_alive():
-                    obj.err_msg += "  - Step Timeout (-w %d set)" % seconds
+                    obj.err_msg += "  - Step Timeout (-t %d set)" % seconds
                     #test_logger.debug('--v-- step did not complete in %d seconds(-t option set) --v--'%seconds)
                     obj.result = False
                     return False
@@ -477,7 +502,7 @@ class TestStep:
                 if tf:
                     ret = func(*args, **kwargs)
                     obj.result = not ret
-                    #test_logger.debug('--v-- reverse the result (due to -x option set) --v--')
+                    test_logger.debug('    vvv  reverse the result (due to -x option set)  vvv')
                     obj.err_msg += '   - Original result: %r (-x option set) ' % ret
                     return obj.result
                 else:
@@ -496,7 +521,8 @@ class TestStep:
                     ret = func(*args, **kwargs)
                     if not ret:
                         obj.result = not ret
-                        test_logger.warn('--v-- condition not met (pass due to -w option set) --v--')
+                        #test_logger.warn('--v-- condition not met (pass due to -w option set) --v--')
+                        obj.warn_msg = '  ^^^  condition not met (pass due to -w option set)  ^^^ '
                     return obj.result
                 else:
                     return func(*args, **kwargs)
@@ -514,7 +540,7 @@ class TestStep:
                 zzz = end_time - time.time()
                 if zzz > 0: time.sleep(zzz)
                 obj.err_msg += '  - sleep %d seconds (-d %d set)' % (zzz, seconds)
-                #test_logger.debug('--v-- sleep %d seconds (due to -d option set) --v--'%zzz)
+                # test_logger.debug('--v-- sleep %d seconds (due to -d option set) --v--'%zzz)
                 obj.result = ret
                 return ret
 
@@ -590,3 +616,4 @@ addStepOption('faildesc', 'f', str, _fail_desc, 'xfail')
 
 check = step
 checks = steps
+s = steps
